from .toolbox import timetools, color, number_padding
from . import soul_work
from . import storage
from . import toolbox
from . import level
from . import tasks

def window(from_term: bool = False):
    header = toolbox.center("《 STATUS 》", "—")
    print(f"\n{header}")
    
    user     = storage.load_data(storage.USER_FILE)
    birthday = user.get("Birthday")
    age      = "Unknown"
    if birthday: age = timetools.get_age(birthday)
    
    stat        = level.Stat()
    lvl, blocks = stat.get_stat_blocks()
    rank        = stat.get_rank(lvl)
    title       = stat.get_title(rank)
    name        = user.get("Name") or "Unknown"
    money       = soul_work.coffer(getter=True)
    
    toolbox.spacer(2)
    print(f" {color('Name:', 'cyan')} {name}")
    print(f"  {color('Age:', 'cyan')} {age}")
    print(f"{color('Level:', 'cyan')} {lvl}")
    print(f"{color('Title:', 'cyan')} {title}")
    print(f"{color('Money:', 'cyan')} ${money:.2f}")
    
    if not blocks:
        toolbox.no_data("No stats available yet")
        toolbox.underline()
        if from_term: print()
        return
    
    toolbox.spacer(2)        
    for block in blocks:
        stat =(f"{color(block['Short'], 'cyan')}  "
             + f"{block['Weight']}   {block['Bar']}   "
             + f"{block['Rank']}")
        print(toolbox.center(stat, fixed=41), "\n") 
    
    missions = tasks.get_missions()
    if not missions:
        toolbox.underline()
        return
    
    toolbox.spacer(2)       
    MISSIONS = toolbox.center("《 MISSIONS 》")
    print(color(MISSIONS, "green"))
    print()
    for index, mission in enumerate(missions):
        tasks.render_mission_card(mission, index)
    
    toolbox.underline()
    if from_term: print()



from tuikit.listools import flatten, normalize
from tuikit.logictools import any_in
from pathlib import Path
import sqlite3 as sql
import json
import os

# Dynamically resolve the main folder
DATA_DIR = Path(__file__).parent / "data"

# Load data safely from a file
def load_data(file:str) -> list | dict:
    file_path = DATA_DIR / file
    try:
        with open(file_path, "r") as f:
            return json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        return []

def load_table(table:str, cur:sql.Cursor, _type:int = 1
              ) -> list[dict]:
    data = []
    for row in cur.execute(f"SELECT * FROM {table}"):
        entry = {}
        if _type == 1:            
            entry["Timestamp"]     = row[0]
            entry["Task"]          = row[1]
            entry["Focus"]         = row[2]
            entry["Fuzziness"]     = row[3]
            entry["Time spent"]    = row[4]
            entry["Quality range"] = eval(row[5])
        elif _type == 2:
            entry["Timestamp"] = row[0]
            entry["Question"]  = jsonsql._sql_repr(row[1])
            entry["Answer"]    = jsonsql._sql_repr(row[2])
            entry["Theme"]     = jsonsql._sql_repr(table)
            entry["Time"]      = float(row[3])
        data.append(entry)
    
    return data

def load_table_as_list(file:str, table:str) -> list[dict]:
    db_name = file.replace(".json", ".db")
    if not os.path.exists(DATA_DIR / db_name): 
        jsonsql.migrate_to_sql(file)    
    data = load_data(file)
    db   = sql.connect(str(DATA_DIR / db_name))
    cur  = db.cursor()
    if "log_data" in file: return load_table(table, cur, 1)
    elif "reflec" in file: return load_table(table, cur, 2)

def get_tables(file:str, repair:bool = True) -> list[str]:
    db_name = file.replace(".json", ".db")
    if not os.path.exists(DATA_DIR / db_name): 
        jsonsql.migrate_to_sql(file)        
    db = sql.connect(str(DATA_DIR / db_name))
    cr = db.cursor()
    tl = cr.execute(f"SELECT tbl_name FROM sqlite_master")
    if not repair: return normalize(flatten(tl))
    return normalize(flatten(tl), use=jsonsql._sql_repr)

class JsonSQL:
    @staticmethod
    def _sql_clean(*args, table:bool = True) -> str:
        if not table:
            if len(args) == 1:
                a = args[0].replace("'m", "’m")
                return args[0].replace("'s", "’s")
            a = args[0].replace("'", "’")
            b = args[1].replace("'", "’")
            return a, b
        
        args = args[0]   
        args = args.replace("&", "and")
        args = args.replace(" ", "_")
        return args.replace("-", "–")

    @staticmethod
    def _sql_repr(text:str) -> str:
        text = text.replace("_and_", " & ")
        text = text.replace("–", "-")
        text = text.replace("_", " ")
        return text

    def add(self, cur: sql.Cursor, table:str, stamp:str, 
            **kw: dict) -> None:
        table = self._sql_clean(table)
        rows  = cur.execute(f"SELECT * FROM {table}")
        keys  = [repr(row[0]) for row in rows]
        if stamp in keys: return
        if len(kw) > 3:
            task, focus, fuzz, durata, qlty = kw.values()
            task = self._sql_clean(task, table=False)
            cur.execute(f"""INSERT INTO {table} VALUES
                        ({stamp}, {task}, {focus}, {fuzz}, 
                         {durata}, '{qlty}')""")
        else:
            qstn, answer, time = kw.values()
            qstn, answer = self._sql_clean(qstn, answer,
                           table=False)
            cur.execute(f"""INSERT INTO {table} VALUES
                        ({stamp}, {repr(qstn)}, 
                         {repr(answer)}, {time})""")
    
    def migrate_log_data(self, data:list[dict], 
                         db:sql.Connection, cur:sql.Cursor
                         ) -> None:  
        tables  = []
        for entry in data:
            cat = entry["Category"]
            if cat not in tables: tables.append(cat)
    
        for table in tables:
            try: cur.execute(f"""CREATE TABLE {table}
                    (timestamp text, task text,focus real, 
                     fuzziness real, duration real, 
                     quality_range text,
                     CONSTRAINT pk_{table.lower()} PRIMARY 
                     KEY (timestamp))""")
            except sql.OperationalError: pass
    
        for entry in data:
            table  = entry["Category"]
            stamp  = repr(entry["Timestamp"])
            task   = repr(entry["Task"])
            focus  = repr(entry["Focus"])
            fuzz   = repr(entry["Fuzziness"])
            durata = repr(entry["Time spent"])
            qlty   = repr(entry["Quality range"])
            self.add(cur, table, stamp, task=task, 
                     focus=focus, fuzz=fuzz, durata=durata, 
                     qlty=qlty)

    def migrate_reflections(self, data:list[dict], 
                         db:sql.Connection, cur:sql.Cursor
                         ) -> None:
        tables  = []
        for entry in data:
            cat = entry["Theme"]
            if cat not in tables: tables.append(cat)
    
        for table in tables:
            table = self._sql_clean(table)
            try: cur.execute(f"""CREATE TABLE {table}
                    (timestamp text, question text, 
                     answer text, time text,
                     CONSTRAINT pk_{table.lower()} PRIMARY 
                     KEY (timestamp))""")
            except sql.OperationalError: pass
    
        for entry in data:
            table  = entry["Theme"]
            stamp  = repr(entry["Timestamp"])
            qstn   = entry["Question"]
            answer = entry["Answer"]
            time   = repr(entry["Time"])
            self.add(cur, table, stamp, qstn=qstn, 
                     answer=answer, time=time)  

    def migrate_to_sql(self, file:str, 
                       data:list[dict]|None = None)-> None:
        db_name = file.replace(".json", ".db")
        path    = DATA_DIR / db_name
        if not os.path.exists(path): data = load_data(file)        
        db  = sql.connect(str(path))
        cur = db.cursor()
        
        if "log_data" in file: 
            self.migrate_log_data(data, db, cur)
        elif "reflect" in file: 
            self.migrate_reflections(data, db, cur)
        
        db.commit()
        db.close()
    
jsonsql = JsonSQL() 



"""
This module brings philosophy and self-reflection into your system:

- Gratitude: Practicing being thankful
           - To be thankful of what is happening or 
             has happened in ones life
           - To avoid taking the good for granted.
- Extra Services Ledger: Tracks when you go above 
                         and beyond.
- Coffer: Your personal bank — all money movements are 
          logged here.
- Almanack: A personal Bible, containing:
          - Your Chief Aim
          - Your Ten Commandments — principles to 
            abide by for becoming a "definite" person.
These additions make the system not just about productivity, but about transformation.
"""
from .toolbox import timetools, choose_from, main_menu
from .toolbox import color, label, variables, warning
from .toolbox import underline, no_data, list_options
from .toolbox import choose, money, center, ExitError
from .toolbox import percent_colored
from .debrief import debrief_review
from operator import itemgetter 
from datetime import datetime
from . import slight_edge
from . import storage
from . import level
import shutil
import time

def get_name():
    user = storage.load_data(storage.USER_FILE)
    try: return user["Name"]
    except (KeyError, TypeError): return "Unknown"

def menu():
    header = center("《 Soul Work 》", line="—")
    print(f"\n{header}\n")
    
    options = {
        "Daily gratitude": gratitude,
        "Slight edge": slight_edge.menu,
        "Extra services ledger": extra_service_ledger,
        "Coffer": coffer,
        "Almanack": almanack
    }
    
    list_options(options)    
    choose(options, src=menu)

def gratitude():
    header = center(" Daily Gratitude ")
    print(f"\n{color(header, 'green')}\n")
    
    start = time.time()    
    print(color("I'm grateful for", underule=1)+":")
    try: graces = [input(color(">>> ", "magenta"))
                  for _ in range(5)]
    except (EOFError, KeyboardInterrupt) as e: 
        if isinstance(e, KeyboardInterrupt): print()
        main_menu()
        return
    
    end      = time.time()
    duration = round((end - start) / 3600, 3)
    print()
    
    tags = label([
        "Depth level (1–10): ", 
        "Uncertainty in depth (±): "
    ])
    
    focus, fuzziness = variables(tags)
        
    focus     = max(0, min(focus,     10))
    fuzziness = max(0, min(fuzziness,  2))

    q_min   = max( 0, (focus - fuzziness) * duration)
    if q_min > 10: q_min = 10
    q_max   = min(10, (focus + fuzziness) * duration)
    if q_max <  0: q_max =  0
    q_range = [q_min, q_max]

    daily_grace = {
        "Timestamp": datetime.now().isoformat(),
           "Graces": graces,
             "Time": time.time()
    }
    
    session = {
            "Timestamp": datetime.now().isoformat()  ,
                 "Task": "Meditation (via gratitude)",
             "Category": "Meditation"                ,
                "Focus": focus                       ,
            "Fuzziness": fuzziness                   ,
           "Time spent": duration                    ,
        "Quality range": [round(q,2) for q in q_range]
    }
    
    storage.save_entry(storage.GRACES, daily_grace)
    storage.save_entry(storage.DATA_FILE, session)
    slight_edge.ballot(1)
    level.Stat("Meditation").level_up()
    footer = center(
        " ✔ Gratitude Session Logged Successfuly! ",
        line="—")
    print(f"\n{footer}\n")

def extra_service_ledger():
    header = center(" Extra Services Ledger ")
    print(f"\n{color(header, 'green')}\n")
    
    tags = label([
        "Service rendered to: ", 
        "Nature of service: "
    ])
    
    options = ["Log entry", "View rendered services"]
    list_options(options, "Choose entry")
    choice = choose(options, 1, extra_service_ledger)
    print()
    if choice == 1:
        header = center(" Viewing rendered services ")
        print(f"{color(header, 'green')}\n")
        debrief_review(1)
        return
    elif choice is None: return
    
    list_options(guide="Enter extra service rendered")    
    try:
        service = {
            "Timestamp": datetime.now().isoformat(),
            "Entry": {
                "Subject": input(tags[0]),
                "Service": input(tags[1])
            },
           "Time": time.time()
        }
    except (EOFError, KeyboardInterrupt) as e: 
        if isinstance(e, KeyboardInterrupt): print()
        main_menu()
        return
    
    storage.save_entry(storage.SERVICES, service)
    slight_edge.ballot(1)
    footer = center(
             " ✔ Extra Service Rendered Logged ", 
             line="—")
    print(f"\n{footer}")

def coffer(getter=False, view=False, cashflow=False):
    file = storage.COFFER
    bank = storage.load_data(file)
    name = get_name()
    if not bank:
        bank = {
            "Main": {
                "Name": name,
                "Amount": 0,
                "Opened": datetime.now().isoformat()
            },
            "Transactions": []
        }
        storage.save_data(file, bank)
      
    coffers = bank["Main"]["Amount"]
    if getter: return coffers
    
    history = bank["Transactions"]
    tags = label([
        "Enter amount to deposit: ", 
        "Enter amount to withdraw: ",
        "Balance:"
    ])
    
    def deposit():
        print()
        try:
            amount = float(input(tags[0]))
            if amount <= 0: raise ValueError
        except BaseException as e:
            if isinstance(e, ExitError()):
                if not isinstance(e, EOFError): print()
                main_menu()
                return
            print()
            underline()
            return
            
        bank["Main"]["Amount"] += amount
        entry = {
            "Timestamp": datetime.now().isoformat(),
                 "Type": "Deposit",
               "Amount": amount,
                 "Time": time.time()
        }
        history.append(entry)
        storage.save_data(file, bank)
        footer = center(
           f" ${amount:.2f} Deposited Successfully! ","—")
        print(f"\n{footer}")
    
    def withdraw():
        print()
        try:
            amount = float(input(tags[1]))
            if 0 <= amount > bank["Main"]["Amount"]:
                raise ValueError
        except BaseException as e:
            if isinstance(e, ExitError()):
                if not isinstance(e, EOFError):print()
                main_menu()
                return
            print()
            if amount > bank["Main"]["Amount"]:
                warning("Can't withdraw more than "
                       +"you have.\n")
            underline()
            return
            
        bank["Main"]["Amount"] -= amount
        entry = {
            "Timestamp": datetime.now().isoformat(),
            "Type": "Withdrawal",
            "Amount": amount,
            "Time": time.time()
        }
        history.append(entry)
        storage.save_data(file, bank)
        footer = center(
            f" ${amount:.2f} Withdrawn Successfully! ",
            line="—")
        print(f"\n{footer}")
    
    def account_details():
        def get_flow(_type:int = 1) -> str:
            total_inflow  = 0
            total_outflow = 0
            months        = []
            for entry in history:
                if entry["Type"] == "Deposit":
                    total_inflow += entry["Amount"]
                else: total_outflow += entry["Amount"]
                month = timetools.timestamp(entry[
                        "Timestamp"]).split()[2]
                if month not in months:
                    months.append(month)
            
            months      = max(1, len(months))
            avg_inflow  = total_inflow  / months
            avg_outflow = total_outflow / months
            cashflow    = avg_inflow - avg_outflow
            
            if _type == 1:
                flow = round(cashflow, 2)
            if _type == 2:
                flow = round(avg_inflow, 2)
            if _type == 3:
                flow = round(avg_outflow * -1, 2)
            
            return percent_colored(flow, '$', '/month')
            
        user     = bank["Main"]["Name"]
        opened   = bank["Main"]["Opened"]
        cashflow = get_flow(_type=1)
        inflow   = get_flow(_type=2)
        outflow  = get_flow(_type=3)
        tags     = label(["User:", "Opened:", "Balance:",
                   "Cashflow:", "Cash-inflow:", 
                   "Cash-outflow:"])
        
        print(tags[0], user)
        print(tags[1], timetools.timestamp(opened))
        print(tags[2], f"${coffers:.2f}")
        print(tags[3], cashflow)
        print(tags[4], inflow)
        print(tags[5], outflow)
        print()
    
    def view_account(from_term=None):
        print()
        if from_term is not None: choice = from_term
        options = [
            "View balance", 
            "View account details",
            "View transentry history"
        ]
        
        if from_term is None:
            list_options(options)
            choice = choose(options, 1, view_account)
        
        if choice is None: return
        elif choice == 0:
            if from_term is None: print()
            print(tags[2], f"${coffers:.2f}\n")
            if from_term is None: underline()
            return
        elif choice == 1:
            if from_term is None: print()
            account_details()
            if from_term is None: underline()
            return
            
        # === Terminal width handling ===
        term_width = shutil.get_terminal_size(
            ).columns
        date_col_width = 11  # Fixed: e.g. "23 Mar 2007"
        
        # Fix alignment for the withdraw column when
        # text size is 15pt and above
        spacer = 1 if term_width == 46 else (2 if 
                 term_width == 43 else 0)
        
        dynamic_cols  = 3
        padding       = dynamic_cols ** 2
        borders       = dynamic_cols
        usable_width  = (term_width 
                      - date_col_width 
                      - padding 
                      - borders)
        dynamic_width = usable_width // dynamic_cols
        extra_space   = usable_width %  dynamic_cols
        deposit_w     = dynamic_width
        withdraw_w    = dynamic_width
        balance_w     = dynamic_width + extra_space
        
        def fmt(text, width):return f"{text:<{width}}"

        # === Table Header ===
        horizontal_line = color(  "—", "magenta")
        vertical_line   = color(" | ", "magenta")
        
        print("\n"+fmt("Date", date_col_width)
             + vertical_line 
             + fmt("Deposit", deposit_w) 
             + vertical_line 
             + fmt("Withdraw", withdraw_w)
             + vertical_line 
             + fmt("Balance", balance_w))
        print(horizontal_line * term_width)
        
        # Trackers to group entries by date
        track = {
              "Current": None,
            "Deposited": None,
            "Withdrawn": None
        }
        
        balance = 0
        data = sorted(history, key=itemgetter("Time")) 
        for pos, entry in enumerate(data):
            stamp = entry["Timestamp"]
            date  = timetools.timestamp(stamp, 0, 1)
            deposited = entry["Amount"] if (
                entry["Type"] == "Deposit") else 0
            withdrawn = entry["Amount"] if (
                entry["Type"] == "Withdrawal") else 0
            
            # If already tracked, reuses values
            # (avoids duplicates)
            if track["Current"] == stamp:
                deposited = track["Deposited"]
                withdrawn = track["Withdrawn"]
                
            try:
                # Peeks ahead to group entries from
                # the same day
                next_up  = data[pos+1]
                ns       = next_up["Timestamp"]
                nxt_date = timetools.timestamp(ns,0,1)
                
                # Skip printing now — wait until 
                # last entry for the date
                if date == nxt_date:
                    deposited += next_up["Amount"] if(
                        next_up["Type"] == 
                        "Deposit" ) else 0
                    withdrawn += next_up["Amount"] if(
                        next_up["Type"] == 
                        "Withdrawal" ) else 0                   
                    track["Current"]   = ns
                    track["Deposited"] = deposited
                    track["Withdrawn"] = withdrawn
                    continue                 
            except IndexError: pass
            
            # Update running balance
            balance += deposited
            balance -= withdrawn
            
            dol = money(deposited, withdrawn, balance)
            deposited, withdrawn, balance_fmt = dol
            print(fmt(date, date_col_width) 
                 + vertical_line 
                 + fmt(deposited, deposit_w) 
                 + vertical_line 
                 + fmt(withdrawn, withdraw_w+spacer)
                 + vertical_line 
                 + fmt(balance_fmt, balance_w))

        underline()
        if from_term is not None: print() 
    
    header = center(" Coffer ")        
    options = {
        "Account": view_account,
        "Deposit": deposit,
        "Withdraw": withdraw
    }
    
    if ( isinstance(    view, bool)
    and  isinstance(cashflow, bool)):
        print(f"\n{color(header, 'green')}\n")
        list_options(options)
        choose(options, src=coffer)
    elif not isinstance(    view, bool):
        view_account(view)
    elif not isinstance(cashflow, bool):
        if cashflow == 0:withdraw()
        if cashflow == 1: deposit()
        print()

def almanack():
    name   = get_name()
    header = center(f" {name}'s Almanack ")
    print(f"\n{color(header, 'green')}\n")
    
    def aim():
        chief_aim = user.get("Aim")
        if chief_aim:
            print(f"\n{chief_aim}\n")
            underline()      
        else: no_data("No Chief Aim Found")
    
    def acts():
        decalogue = user.get("Decalogue")
        if decalogue:
            print()
            list_options(decalogue, 0, 1)
            underline()
        else: no_data("No Commandments Found")
    
    options = {
        "View chief aim": aim,
        "View Decalogue": acts
    }
    
    list_options(options)
    choose(options, src=almanack)



"""
This module is meant to streamline deep reflection
Main functions:
  -        reflect(...)
  - shadow_journal(...)
  - debrief_review(...)
"""
from .toolbox import choose_from, strip_ansi, no_data
from .toolbox import choose, list_options, underline
from .toolbox import wrap_text as wrap, center as c
from .toolbox import warning, label, variables
from .toolbox import color, timetools
from .slight_edge import ballot
from datetime import datetime
from time import time
from . import storage
from . import level
import random

def menu():
    header = c("《 Debrief 》", line="—")
    print(f"\n{header}\n")
        
    options = {
        "Daily reflection": reflect,
        "Shadow work journal": shadow_journal,
        "Debrief review": debrief_review
    }
    
    list_options(options)
    choose(options, src=menu)

def reflect(from_term: bool = False):
    """
Reflect deeply from a series of 5 to 10 questions per session
    """
    line = "—" if from_term else ""
    header = c(" Daily Reflection ", line)
    print(f"\n{color(header, fg='green')}")
    
    # Load the pool of questions from the JSON file
    data = storage.load_data(storage.QUESTIONS)

    start = time()
    
    # Flatten the questions into a single list with 
    # themes
    pool = [{
        "question": question, 
           "theme": theme
    } for theme, questions in data.items()
      for question in questions]
    
    count    = random.randint(5, 10)
    selected = random.sample(pool, count)

    try:
      for n, item in enumerate(selected):
         i = len(f"Q{n+1}. ")
         print()
         print(wrap(f"Q{n+1}. {item['question']}", i))
         reply     = input(color(">>> ", "magenta"))
         timestamp = datetime.now().isoformat()
        
         # Save each reflection as a separate entry
         reflection = {
              "Question": item["question"],
                "Answer": reply           ,
                 "Theme": item["theme"]   ,
             "Timestamp": timestamp       ,
                  "Time": time()
         }        
         storage.save_entry(storage.REFLECTIONS, 
                            reflection)       
    
      end = time()    
      duration = round((end - start) / 3600, 3)    
    
      labels = label([
          "Depth level (1–10): ", 
          "Uncertainty in depth (±): "
      ])    
    
      focus, fuzziness = variables(labels)
    except(KeyboardInterrupt, EOFError, TypeError)as e:
        if     isinstance(e,  EOFError): print()
        if not isinstance(e, TypeError):
            print()
            underline()
        if from_term: print()
        return
    
    focus     = max(0, min(focus,     10))
    fuzziness = max(0, min(fuzziness,  2))

    q_min   = max( 0, (focus - fuzziness) * duration)
    if q_min > 10: q_min = 10
    q_max   = min(10, (focus + fuzziness) * duration)
    if q_max <  0: q_max =  0
    q_range = [q_min, q_max]

    session = {
            "Timestamp": datetime.now().isoformat()  ,
                 "Task": "Daily reflection session"  ,
             "Category": "Meditation"                ,
                "Focus": focus                       ,
            "Fuzziness": fuzziness                   ,
           "Time spent": duration                    ,
        "Quality range": [round(q,2) for q in q_range]
    }
    
    storage.save_entry(storage.DATA_FILE, session)
    
    ballot(1)
    level.Stat("Meditation").level_up()
    done = c(" ✔ Reflection Logged Successfuly! ","—")
    print(f"\n{done}")
    if from_term: print()

def shadow_journal(from_term: bool = False):
    """Reflect on what recently occured to you"""
    line = "—" if from_term else ""
    header = c(" Shadow Work Journal ", line)
    print(f"\n{color(header, fg='green')}")
    
    start = time()
    
    long = [
"How did you react externally (words, body language, tone)?",
"Does this remind you of anything from your past?",
"What belief about yourself or others was exposed?",
"What did you need in that moment but didn't get?",
"How can you view this situation differently now?",
"How do you want to handle similar moments going forward?"
    ]
    
    wrapped = [wrap(q) for q in long]
    prompt = color('>>> ', 'magenta')
    
    questions = [
f"\nWhat happened?\n{prompt}",
f"\nWhat did you feel?\n{prompt}",
f"\n{wrapped[0]}\n{prompt}",
f"\nWhy did it affect you so strongly?\n{prompt}",
f"\n{wrapped[1]}\n{prompt}",
f"\n{wrapped[2]}\n{prompt}",
f"\n{wrapped[3]}\n{prompt}",
f"\n{wrapped[4]}\n{prompt}",
f"\n{wrapped[5]}\n{prompt}"
    ]
    
    try:
        entry = {
         "Trigger Event": input(questions[0])       ,
    "Emotional Response": input(questions[1])       ,
              "Reaction": input(questions[2])       ,
          "Deeper Truth": input(questions[3])       ,
             "Past Echo": input(questions[4])       ,
           "Core Belief": input(questions[5])       ,
                  "Need": input(questions[6])       ,
               "Reframe": input(questions[7])       ,
             "Next Time": input(questions[8])       ,
             "Timestamp": datetime.now().isoformat(),
                  "Time": time()
        }    
    
        end      = time()
        duration = round((end - start) / 3600, 3)
    
        labels = label([
            "Depth level (1–10): ", 
            "Uncertainty in depth (±): "
        ])
    
        focus, fuzziness = variables(labels)    
    except(KeyboardInterrupt, EOFError, TypeError)as e:
        if     isinstance(e,  EOFError): print()
        if not isinstance(e, TypeError):
            print()
            underline()
        if from_term: print()
        return
    
    focus     = max(0, min(focus,     10))
    fuzziness = max(0, min(fuzziness,  2))

    q_min   = max( 0, (focus - fuzziness) * duration)
    if q_min > 10: q_min = 10
    q_max   = min(10, (focus + fuzziness) * duration)
    if q_max <  0: q_max =  0
    q_range = [q_min, q_max]

    session = {
            "Timestamp": datetime.now().isoformat()  ,
                 "Task": "Shadow work"               ,
             "Category": "Meditation"                ,
                "Focus": focus                       ,
            "Fuzziness": fuzziness                   ,
           "Time spent": duration                    ,
        "Quality range": [round(q,2) for q in q_range]
    }
    
    storage.save_entry(storage.JOURNAL, entry)
    storage.save_entry(storage.DATA_FILE, session)
    ballot(1)
    level.Stat("Meditation").level_up()
    done = c(" ✔ Entry Logged Successfully ", "—")
    print(f"\n{done}")
    if from_term: print()

def debrief_review(proxy: bool = False, from_term: 
                   bool = False):
    """
Opens the Debrief Review menu
If proxy=True, it means this was called from another module, so we skip the theme filter to avoid unnecessary nesting

Purpose:
    view logged reflections for analysis (you can 
    copy-paste them to ChatGPT for an in-depth 
    analysis)
    """
    if not proxy:
        line = "—" if from_term else ""
        header = c(' Debrief Review ', line)
        print(f"\n{color(header, 'green')}\n")
    
    options = {
         "View all entries": view_all_entries,
          "Filter by theme": filter_by_theme ,
           "Filter by date": filter_by_date  ,
        "Search by keyword": search_keyword
    }
    
    if proxy:
        options = {
             "View all entries": view_all_entries,
               "Filter by date": filter_by_date  ,
            "Search by keyword": search_keyword
        }
            
    list_options(options)
    
    if proxy: choose(options, 0, debrief_review, 1)
    else    : choose(options, 0, debrief_review)
    
    if from_term: print()

def choose_journal_type() -> tuple:
    """
Prompts the user to choose between viewing Reflections or Shadow Journal entries.

Returns:
  - A string indicating the type ('reflections' or 
    'shadow')
  - The corresponding loaded data
    """
    print()
    types = ["Reflections", "Shadow Journal"]
    list_options(types, "Choose entry type")
    choice = input(color(">>> ", "magenta")).strip()
    if choice == "1":
        return "reflections", storage.load_data(
                storage.REFLECTIONS)
    elif choice == "2":
        return "shadow", storage.load_data(
                storage.JOURNAL)
    
    warning("Invalid choice.")
    return None, []

def view_all_entries(proxy: bool = False):
    """
Displays all entries from either reflections or shadow journals
If proxy=True, it's being called internally (e.g. from another filter), so it skips the journal type selection
    """
    entries = storage.load_data(storage.SERVICES)
    if not proxy:kind, entries = choose_journal_type()
    if not entries: no_data("No entries found")
    if not entries: return
    
    if not proxy: print_entries(entries, kind)
    else        : print_entries(entries)

def filter_by_theme():
    """
Allows user to filter reflection entries by their associated theme
Only applies to reflection-type journals (not shadow work)
    """
    kind, entries = choose_journal_type()
    if not entries: no_data("No entries found")
    if not entries: return

    file   = storage.REFLECTIONS
    themes = storage.get_tables(file, repair=False)

    list_options(storage.get_tables(file), "Choose theme")
    
    try:
        i = int(input(color(">>> ", 'magenta'))) - 1
        if 0 <= i < len(themes):
            print_entries(storage.load_table_as_list(file, 
                          themes[i]), kind)
        else: warning("Invalid choice")
    except EOFError: warning("Invalid input")

def filter_by_date(proxy: bool = False):
    """
Filters journal entries based on a given date
    """
    print()
    entries = storage.load_data(storage.SERVICES)
    
    if not proxy:kind, entries = choose_journal_type()
    if not entries: no_data("No entries found")
    if not entries: return

    date_input = input("\nEnter date (YYYY-MM-DD): "
                 ).strip()
    selected = [e for e in entries if 
                e["Timestamp"].startswith(date_input)]
    if not selected:
        no_data("No entries found for that date")
    else:
        if not proxy:
            print_entries(selected, kind)
            return
        print_entries(selected)

def search_keyword(proxy: bool = False):
    """
Searches journal entries for a user-specified keyword
If proxy=True, the function is being called internally, so it handles data differently.
    """
    print()
    entries = storage.load_data(storage.SERVICES)
    if not proxy:kind, entries = choose_journal_type()
    if not entries: no_data("No entries found")
    if not entries: return

    prompt  = color("Enter keyword to search:","cyan")
    keyword = input(f"\n{prompt} ").lower()
    
    # Match logic varies based on type and proxy status
    def match(entry):
        if kind == "reflections":
            return keyword in entry["Answer"
                   ].lower() or keyword in entry[
                   "Question"].lower()
        elif proxy:
            return keyword in entry["Entry"]["Subject"
                   ].lower() or keyword in entry[
                   "Entry"]["Service"].lower()
        else:
            return any(keyword in str(v).lower() for 
                   k, v in entry.items() if 
                   isinstance(v, str))
    
    selected = [e for e in entries if match(e)]
    if not selected: no_data("No results found")
    else:
        if not proxy:
            print_entries(selected, kind)
            return
        print_entries(selected)

def print_entries(entries: list, kind: str|None=None):
    """
Prints journal entries in a readable format based on entry type.
    kind='reflections': Shows Q&A with theme
         kind='shadow': Shows shadow work prompts
             kind=None: Shows service log style 
                        entries (used in soul_work.py)
    """
    for e in entries:
        print(end="\n\n")
        stamp = timetools.timestamp(e['Timestamp'])
        ago   = timetools.get_time_diff(time(), e['Time'])
        
        labels = label([
            "Date:", "Theme:", "Q:", "A:", 
            "Rendered to:", "Service:"
        ])
        
        print(labels[0], stamp)
        if kind == "reflections":
            question = f'{labels[2]} {e["Question"]}'
            answer   = f'{labels[3]} {e["Answer"]}'            
            print(labels[1], e["Theme"])
            print(wrap(question, 3))
            print(wrap(answer,   3))
        elif not kind:
            subject = f" {e['Entry']['Subject']}"
            service = f" {e['Entry']['Service']}"
            sub     = len(strip_ansi(labels[4]) + 1)
            ser     = len(strip_ansi(labels[5]) + 1)
            print(wrap(f"{labels[4]}{subject}", sub))
            print(wrap(f"{labels[5]}{service}", ser))
        else:
            for k, tag in {
                     "Trigger Event": "Trigger" ,
                "Emotional Response": "Emotion" ,
                          "Reaction": "Reaction",
                      "Deeper Truth": "Truth"   ,
                         "Past Echo": "Past"    ,
                       "Core Belief": "Belief"  ,
                              "Need": "Need"    ,
                           "Reframe": "Reframe" ,
                         "Next Time": "Next"
            }.items():
                legend  = color(f"{tag}:", "cyan")
                content = f"{legend} {e.get(k, '')}"
                print(wrap(content, len(tag)+2))
        print(color("Logged:", "cyan"), ago)
    print(); underline()



from .toolbox import center, color, choose, warning
from .toolbox import make_progress_bar, underline
from .toolbox import list_options, choose_from
from .toolbox import no_data, timetools, label
from .toolbox import visual_width
from datetime import datetime
from . import storage
import shutil

def menu():
    header = center(" The Slight Edge ")
    print(f"\n{color(header, 'green')}\n")
    
    options = {
        "Vote action": ballot,
        "View trajectory": direction
    }
    
    list_options(options)  
    choose(options, src=menu)

def ballot(voting_via_proxy=False):
    data  = storage.load_data(storage.SLIGHT_EDGE)
    today = datetime.now().date()
    def new_day(vote=1):
        entry = {
            "Day": datetime.now().isoformat(),
            "Action": 1,
            "Score": vote,
            "Edge": (vote / 1) * 100
        }
        storage.save_entry(storage.SLIGHT_EDGE, entry)
        
    if voting_via_proxy:
        found = False
        for entry in data:
            if datetime.fromisoformat(entry["Day"]
                                ).date() == today:
                entry["Score"]  += 1
                entry["Action"] += 1
                entry["Edge"] = (entry["Score"] 
                              / entry["Action"]) * 100
                storage.save_data(storage.SLIGHT_EDGE, 
                    data)
                found = True
                break
            
        if not found: new_day()
        return
        
    header = center(" Ballot ")
    print(f"\n{color(header, 'green')}\n")
   
    list_options(["Good","Bad"], "Vote action")
    choice = choose([0, 1], 1, ballot)
    
    if choice is None: return
    
    vote = 1 if choice == 0 else 0

    found = False
    for entry in data:
        if datetime.fromisoformat(entry["Day"]).date(
                                          ) == today:
            entry["Score"]  += vote
            entry["Action"] += 1
            entry["Edge"] = (entry["Score"] 
                          / entry["Action"]) * 100
            storage.save_data(storage.SLIGHT_EDGE, 
                data)
            found = True
            break
            
    if not found: new_day(vote)
    
    footer = center(" Action Voted Successfully ","—")
    print("\n"+footer)

def direction():
    data = storage.load_data(storage.SLIGHT_EDGE)
    if not data:
        no_data("No decision/action votes found")
        underline()
        return
        
    header = center(" Life Trajectory ")
    print(f"\n{color(header, 'green')}\n")   
    
    options = [
        "Overall trajectory", 
        "Daily trajectory",
        "Trajectory graphs"
    ]
    
    list_options(options)    
    choice = choose(options, 1, direction)
    
    if choice is None: return
    elif choice < 2:
        print()
        display(choice)
    else: graph()
    underline()

def display(daily: bool = False):
    data    = storage.load_data(storage.SLIGHT_EDGE)
    actions = sum([entry["Action"] for entry in data])
    score   = sum([entry["Score"] for entry in data])    
    edge    = score / actions * 100
    trajectory = "Headed for success" if (edge >= 50
          ) else "Headed for failure"    
    
    labels = label([
        "Date:", "Trajectory:", "Chance of success:"
    ])
    
    if not daily:
        print(labels[1], trajectory)
        print(f"{labels[2]} {edge:.2f}%\n")
        return

    for entry in data:
        date = timetools.timestamp(entry["Day"], 1)
        trajectory = "Headed for success" if entry[
           "Edge"] >= 50 else "Headed for failure" 
        print(labels[0], date)
        print(labels[1], trajectory)
        print(f"{labels[2]} {entry['Edge']:.2f}%\n")

def graph():
    header = center(" Edge Trajectory Graphs ")
    print(f"{color(header, 'green')}\n")

    options = {
        "Bar Graph (Edge % per day)": edge_bar_graph,
        "Trendline": edge_trendline
    }

    list_options(options)
    choose(options)

def edge_bar_graph(limit: int = 7):
    print()
    data  = storage.load_data(storage.SLIGHT_EDGE)
    limit = min(limit, len(data))
    for entry in data[-limit:]:
        date = timetools.timestamp(entry["Day"], 0, 1)
        edge = entry["Edge"]
        bar  = make_progress_bar(edge / 100)
        print(f"\n{date[:-5]} {bar} {edge:>5.1f}%")
    print()

def edge_trendline(limit: int = 7):
    print()    
    data   = storage.load_data(storage.SLIGHT_EDGE)
    limit  = min(limit, len(data))
    values = [round(entry["Edge"]) for entry in 
             data[-limit:]]
    dates  = [timetools.timestamp(e["Day"], 0, 1)[:2] 
             for e in data[-limit:]]  # day only

    for y in range(100, -1, -5):
        row = f"{y:>3} ┤"
        for pos, v in enumerate(values):
            if v == 100: point = color("●", "magenta")
            elif 100 > v >= 75:
                point = color("●", "green")
            elif 75 > v > 30:
                point = color("●", "yellow")
            else: point = color("●", "red")
            filler = color("○", "gray")
            row += f"  {point} "  if (y - 5 < v <= y 
            ) else f"  {filler} "
        print(center(row))

    term_width  = shutil.get_terminal_size((80, 
                  20)).columns
    display_len = 28 if term_width > 49 else 25
    f           = 28 if term_width > 55 else 26
    print(center(" └" + "─" * display_len, fixed=f))
    print(center("   " + "  ".join(dates), fixed=f))
    print()


from pathlib import Path
from .jsonsql import *
import json
import os

STAT_MAP    = 'stat_map.json'
USER_FILE   = 'user_data.json'
DATA_FILE   = 'log_data.json'
LEVEL_FILE  = 'level_data.json'
S_LVL       = 'stat_levels.json'
TASK_FILE   = 'mission_data.json'
QUESTIONS   = 'questions.json'
REFLECTIONS = 'reflections.json'
JOURNAL     = 'journal.json'
NOTEBOOK    = 'notebook.json'
TRASHCAN    = 'trashcan.json'
SLIGHT_EDGE = 'slight_edge.json'
SERVICES    = 'services_ledger.json'
GRACES      = 'gratitudes.json'
COFFER      = 'coffer.json'

# Load stats map
def load_map() -> dict:
    STAT_MAP_FILE = DATA_DIR / STAT_MAP
    if not STAT_MAP_FILE.exists():
        return {"Renames": {}, "Combinations": {}}
    with open(STAT_MAP_FILE, "r") as f:
        return json.load(f)

# Save data to a file (overwrites)
def save_data(file:str, data: list|dict) -> None:
    file_path = DATA_DIR / file
    with open(file_path, "w") as f:
        json.dump(data, f, indent=2)

# Add one entry to a file (append style)
def save_entry(file:str, entry:dict) -> None:
    data = load_data(file)
    data.append(entry)
    save_data(file, data)
    
    if not any_in("log_data", "reflect", eq=file): return
        
    #append to sqlite database
    jsonsql.migrate_to_sql(file, data=[entry])

def delete(file:str) -> None: save_data(DATA_DIR/file, {})



from .toolbox import no_data, center, timetools, label
from .toolbox import color, underline, choose 
from .toolbox import list_options
from datetime import datetime
from . import storage
from . import level
import time

def menu():
    print(f"\n{center('《 Notebook 》', '—')}\n")
    
    options = {
        "Write a new page": write,
        "Read a page": read,
        "Edit a page": edit,
        "Delete a page": del_page,
        "Trashcan": trashcan
    }
    list_options(options, "Choose action")   
    choose(options, src=menu)

def write():
    header = center(" New Page ")
    print(f"\n{color(header, 'green')}\n")
    
    labels = label(["Enter Title: ", "Write notes:"])
    
    start = time.time()
    page  = {
        "Title": input(f"{labels[0]}") or "Untitled",
        "Content": input(f"{labels[1]}\n\n"),
        "Timestamp": datetime.now().isoformat(),
        "Time": time.time()
    }
    end      = time.time()
    duration = (end - start) / 3600
    
    session = {
        "Timestamp": datetime.now().isoformat(),
        "Task": "Writing notes",
        "Category": "Writing",
        "Focus": 5,
        "Fuzziness": 0,
        "Time spent": duration,
        "Quality range": [5 * duration, 5 * duration]
    }
    
    storage.save_entry(storage.NOTEBOOK, page)
    storage.save_entry(storage.DATA_FILE, session)
    level.Stat("Writing").level_up()
    footer = center(" ✔ Page Saved Successfuly! ","—")
    print(f"\n{footer}")

def read(page=None):
    labels = label(["Title:", "Edited:"])    
    if page:
        stamp = timetools.timestamp(page["Timestamp"])
        print(f"\n{labels[0]} {page['Title']}\n")
        print(f"{page['Content']}\n")
        print(f"{labels[1]} {stamp}\n")
        underline()
        return
        
    notebook = storage.load_data(storage.NOTEBOOK)
    if not notebook:
        no_data("No notes written yet")
        underline()
        return
                      
    print(f"\n{color('Choose page', underule=1)}"+":")
    for pos, page in enumerate(notebook):
        print(f"{pos+1}.", page["Title"])  
    
    choice = choose(notebook, 1, read)
    if choice is None: return        
    page  = notebook[choice]
    stamp = timetools.timestamp(page["Timestamp"])
    print(f"\n{labels[0]} {page['Title']}\n")
    print(f"{page['Content']}\n")
    print(f"{labels[1]} {stamp}\n")
    underline()

def edit():    
    notebook = storage.load_data(storage.NOTEBOOK)
    if not notebook:
        no_data("No notes written yet")
        underline()
        return
    
    print(f"\n{color('Choose page', underule=1)}"+":")
    for pos, page in enumerate(notebook):
        print(f"{pos+1}.", page["Title"])  
    
    choice = choose(notebook, 1, read)
    if choice is None: return
    
    edited = notebook[choice]    
    labels = label([
        "Enter new title: ", 
        f"Add more notes:\n\n{edit['Content']}\n"
    ])
    
    print()
    list_options(["Edit title", "Edit content"],
                  "Choose action")
    choice = choose([0, 1], 1, edit)
    if choice is None: return
    elif choice == 0:
        edited["Title"] = input(labels[0])
    elif choice == 1:
        edited["Content"] += "\n"+input(labels[1])
    
    edited["Timestamp"] = datetime.now().isoformat()
    storage.save_data(storage.NOTEBOOK, notebook)
    foot = center(" ✔ Page Edited Successfuly! ", "—")
    print(f"\n{foot}")

def del_page():
    notebook = storage.load_data(storage.NOTEBOOK)
    if not notebook:
        no_data("No notes written yet")
        underline()
        return
    
    print(f"\n{color('Choose page', underule=1)}"+":")
    for pos, page in enumerate(notebook):
        print(f"{pos+1}.", page["Title"])  
    
    choice = choose(notebook, 1, del_page)
    if choice is None: return        
    
    page = notebook[choice]    
    text = color("\nAre you sure you want to delete "
         + f"{page['Title']}? [y/n]: ", "red") 
    confirmation = input(text)[0].lower().strip()
    
    if confirmation != "y":
        footer = center(" ✔ Page Not Deleted ", "—")
        print(f"\n{footer}\n")
        return
        
    page["Position"] = choice
    storage.save_entry(storage.TRASHCAN, page)
    del notebook[choice]
    storage.save_data(storage.NOTEBOOK, notebook)
    footer = center(" ✔ Page Moved to Trashcan ", "—")
    print(f"\n{footer}")
    
def trashcan():
    trash = storage.load_data(storage.TRASHCAN)
    if not trash:
        no_data("No deleted pages found")
        underline()
        return
        
    header = center(" Trashcan ")
    print(f"\n{color(header, 'green')}\n")
        
    notebook = storage.load_data(storage.NOTEBOOK)
    
    print(f"{color('Choose page', underule=1)}"+":")
    for pos, page in enumerate(trash):
        print(f"{pos+1}.", page["Title"])
    
    index = choose(trash, 1, trashcan)
    if index is None: return        
    
    page = trash[index]
    
    print()
    options = ["View page", "Restore page"]   
    list_options(options, "Choose action")    
    choice = choose(options, 1, trashcan)
    if choice is None: return 
    
    if choice == 0:
        read(page)
        return
        
    notebook.insert(page["Position"], page)
    del trash[index]
    storage.save_data(storage.TRASHCAN, trash)
    storage.save_data(storage.NOTEBOOK, notebook)
    foot =center(" ✔ Page Restored Successfuly! ","—")    
    print(f"\n{foot}")